"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDefaultColumnTypeForDecorator = getDefaultColumnTypeForDecorator;
exports.convertArgumentToColumnType = convertArgumentToColumnType;
exports.convertTsTypeToColumnType = convertTsTypeToColumnType;
exports.convertTypeToColumnType = convertTypeToColumnType;
exports.isTypesEqual = isTypesEqual;
exports.typeToString = typeToString;
const utils_1 = require("@typescript-eslint/utils");
const typescript_1 = require("typescript");
const treeTraversal_js_1 = require("./treeTraversal.js");
// From: https://github.com/typeorm/typeorm/blob/master/src/driver/types/ColumnTypes.ts
const booleanLike = ['boolean', 'bool'];
const numberLike = [
    'bigint',
    'dec',
    'decimal',
    'fixed',
    'int',
    'int2',
    'int4',
    'int8',
    'integer',
    'mediumint',
    'number',
    'numeric',
    'smalldecimal',
    'smallint',
    'tinyint',
    'dec',
    'decimal',
    'double precision',
    'double',
    'fixed',
    'float',
    'number',
    'numeric',
    'real',
    'smalldecimal',
];
const stringLike = [
    'character varying',
    'varying character',
    'char varying',
    'nvarchar',
    'national varchar',
    'character',
    'native character',
    'varchar',
    'char',
    'nchar',
    'national char',
    'varchar2',
    'nvarchar2',
    'alphanum',
    'shorttext',
    'string',
    'text',
    'uuid',
];
const dateLike = [
    'date',
    'datetime',
    'datetime2',
    'datetimeoffset',
    'interval day to second',
    'interval year to month',
    'interval',
    'seconddate',
    'smalldatetime',
    'time with time zone',
    'time without time zone',
    'time',
    'timestamp with local time zone',
    'timestamp with local time zone',
    'timestamp with time zone',
    'timestamp without time zone',
    'timestamp',
    'timestamptz',
    'timetz',
    'year',
];
function convertTypeOrmToColumnType(arg) {
    if (booleanLike.includes(arg)) {
        return 'boolean';
    }
    if (numberLike.includes(arg)) {
        return 'number';
    }
    if (stringLike.includes(arg)) {
        return 'string';
    }
    if (dateLike.includes(arg)) {
        return 'Date';
    }
    return 'unknown';
}
function getDefaultColumnTypeForDecorator(column) {
    switch (column) {
        case 'PrimaryColumn':
        case 'PrimaryGeneratedColumn':
        case 'VersionColumn':
            return { type: 'integer', nullable: false };
        case 'CreateDateColumn':
        case 'UpdateDateColumn':
            return { type: 'datetime', nullable: false };
        case 'DeleteDateColumn':
            return { type: 'datetime', nullable: true };
        case 'Column':
        default:
            return {};
    }
}
function convertArgumentToColumnType(column, args) {
    var _a, _b;
    const parsed = args.reduce((prev, arg) => {
        switch (arg.type) {
            case utils_1.AST_NODE_TYPES.ObjectExpression:
                return Object.assign(Object.assign({}, prev), (0, treeTraversal_js_1.parseObjectLiteral)(arg));
            case utils_1.AST_NODE_TYPES.Literal:
                if (typeof arg.value === 'string') {
                    return Object.assign(Object.assign({}, prev), { type: arg.value });
                }
                return prev;
            default:
                return prev;
        }
    }, getDefaultColumnTypeForDecorator(column));
    return {
        columnType: parsed.type && !parsed.transformer
            ? convertTypeOrmToColumnType(parsed.type)
            : 'unknown',
        nullable: (_a = parsed.nullable) !== null && _a !== void 0 ? _a : false,
        literal: false,
        array: (_b = parsed.array) !== null && _b !== void 0 ? _b : false,
    };
}
function convertTsTypeToColumnType(arg, checker) {
    switch (arg.kind) {
        case typescript_1.SyntaxKind.TemplateLiteralType:
        case typescript_1.SyntaxKind.StringKeyword:
            return { columnType: 'string', nullable: false, literal: false, array: false };
        case typescript_1.SyntaxKind.NumberKeyword:
        case typescript_1.SyntaxKind.BigIntKeyword:
            return { columnType: 'number', nullable: false, literal: false, array: false };
        case typescript_1.SyntaxKind.BooleanKeyword:
            return { columnType: 'boolean', nullable: false, literal: false, array: false };
        case typescript_1.SyntaxKind.TypeReference:
            if ((0, typescript_1.isTypeReferenceNode)(arg)) {
                const symbol = checker.getTypeAtLocation(arg.typeName).getSymbol();
                if ((symbol === null || symbol === void 0 ? void 0 : symbol.getName()) === 'Date') {
                    return { columnType: 'Date', nullable: false, literal: false, array: false };
                }
            }
            break;
        case typescript_1.SyntaxKind.LiteralType: {
            const literal = arg;
            switch (literal.literal.kind) {
                case typescript_1.SyntaxKind.NullKeyword:
                    return { columnType: 'unknown', nullable: true, literal: false, array: false };
                case typescript_1.SyntaxKind.StringLiteral:
                    return { columnType: 'string', nullable: false, literal: true, array: false };
                case typescript_1.SyntaxKind.NumericLiteral:
                    return { columnType: 'number', nullable: false, literal: true, array: false };
                case typescript_1.SyntaxKind.TrueKeyword:
                case typescript_1.SyntaxKind.FalseKeyword:
                    return { columnType: 'boolean', nullable: false, literal: true, array: false };
                default:
                    break;
            }
            break;
        }
        case typescript_1.SyntaxKind.ArrayType: {
            const array = arg;
            const item = convertTsTypeToColumnType(array.elementType, checker);
            return Object.assign(Object.assign({}, item), { array: true });
        }
        case typescript_1.SyntaxKind.UnionType: {
            const union = arg;
            return union.types.reduce((acc, currentType) => {
                const current = convertTsTypeToColumnType(currentType, checker);
                return {
                    columnType: current.columnType !== 'unknown' ? current.columnType : acc.columnType,
                    nullable: current.nullable || acc.nullable,
                    literal: current.literal || acc.literal,
                    array: current.array || acc.array,
                };
            }, {
                columnType: 'unknown',
                nullable: false,
                literal: false,
                array: false,
            });
        }
        default:
            return { columnType: 'unknown', nullable: false, literal: false, array: false };
    }
    return { columnType: 'unknown', nullable: false, literal: false, array: false };
}
function convertTypeToColumnType(arg) {
    switch (arg.type) {
        case utils_1.AST_NODE_TYPES.TSStringKeyword:
            return { columnType: 'string', nullable: false, literal: false, array: false };
        case utils_1.AST_NODE_TYPES.TSBigIntKeyword:
        case utils_1.AST_NODE_TYPES.TSNumberKeyword:
            return { columnType: 'number', nullable: false, literal: false, array: false };
        case utils_1.AST_NODE_TYPES.TSBooleanKeyword:
            return { columnType: 'boolean', nullable: false, literal: false, array: false };
        case utils_1.AST_NODE_TYPES.TSNullKeyword:
            return { columnType: 'unknown', nullable: true, literal: false, array: false };
        case utils_1.AST_NODE_TYPES.TSTypeReference:
            if (arg.typeName.type === utils_1.AST_NODE_TYPES.Identifier && arg.typeName.name === 'Date') {
                return { columnType: 'Date', nullable: false, literal: false, array: false };
            }
            return { columnType: 'unknown', nullable: false, literal: false, array: false };
        case utils_1.AST_NODE_TYPES.TSUnionType:
            return arg.types.reduce((acc, currentNode) => {
                const current = convertTypeToColumnType(currentNode);
                if (current) {
                    return {
                        columnType: current.columnType !== 'unknown'
                            ? current.columnType
                            : acc.columnType,
                        nullable: current.nullable || acc.nullable,
                        literal: current.literal || acc.literal,
                        array: current.array || acc.array,
                    };
                }
                return acc;
            }, {
                columnType: 'unknown',
                nullable: false,
                literal: false,
                array: false,
            });
        case utils_1.AST_NODE_TYPES.TSLiteralType: // Literal type
            switch (arg.literal.type) {
                case utils_1.AST_NODE_TYPES.Literal: {
                    const literalType = typeof arg.literal.value;
                    if (['string', 'number', 'boolean'].includes(literalType)) {
                        return {
                            columnType: literalType,
                            nullable: false,
                            literal: true,
                            array: false,
                        };
                    }
                    return { columnType: 'unknown', nullable: false, literal: true, array: false };
                }
                case utils_1.AST_NODE_TYPES.TemplateLiteral:
                    return { columnType: 'string', nullable: false, literal: true, array: false };
                default:
                    return { columnType: 'unknown', nullable: false, literal: true, array: false };
            }
        case utils_1.AST_NODE_TYPES.TSArrayType: {
            const item = convertTypeToColumnType(arg.elementType);
            return Object.assign(Object.assign({}, item), { array: true });
        }
        // TODO: handles these types too
        case utils_1.AST_NODE_TYPES.TSObjectKeyword: // Object type
        case utils_1.AST_NODE_TYPES.TSAnyKeyword: // Unknown types
        case utils_1.AST_NODE_TYPES.TSUndefinedKeyword:
        case utils_1.AST_NODE_TYPES.TSUnknownKeyword:
        default:
            return { columnType: 'unknown', nullable: false, literal: false, array: false };
    }
}
function isTypesEqual(toType, tsType) {
    // If either is unknown, we only check the nullability
    if (toType.columnType === 'unknown' || tsType.columnType === 'unknown') {
        return toType.nullable === tsType.nullable;
    }
    // Dates can be parsed into strings too
    if (toType.columnType === 'Date' &&
        tsType.columnType === 'string' &&
        toType.nullable === tsType.nullable &&
        toType.array === tsType.array) {
        return true;
    }
    // Otherwise just check field equality
    return (toType.columnType === tsType.columnType &&
        toType.nullable === tsType.nullable &&
        toType.array === tsType.array);
}
function typeToString(column, { literal, columnType: tsColumnType }) {
    // If column type is unknown, we fall back to the TypeScript column type
    const columnType = column.columnType !== 'unknown' ? column.columnType : tsColumnType;
    // If type is unknown or literal, we don't suggest change
    if (columnType === 'unknown' || literal) {
        return undefined;
    }
    return `${columnType}${column.array ? '[]' : ''}${column.nullable ? ' | null' : ''}`;
}
